@@ -1,22 +1,8 @@
-// Copyright (c) 2017 fd developers
-// Licensed under the Apache License, Version 2.0
-// <LICENSE-APACHE or http://www.apache.org/licenses/LICENSE-2.0>
-// or the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>,
-// at your option. All files in the project carrying such
-// notice may not be copied, modified, or distributed except
-// according to those terms.
-
-use crate::exec;
-use crate::exit_codes::ExitCode;
-use crate::fshelper;
-use crate::internal::{opts::FdOptions, osstr_to_bytes, MAX_BUFFER_LENGTH};
-use crate::output;
-
 use std::borrow::Cow;
-use std::error::Error;
 use std::ffi::OsStr;
+use std::fs::{FileType, Metadata};
 use std::io;
-use std::path::PathBuf;
+use std::path::{Path, PathBuf};
 use std::process;
 use std::sync::atomic::{AtomicBool, Ordering};
 use std::sync::mpsc::{channel, Receiver, Sender};
@@ -24,10 +10,18 @@ use std::sync::{Arc, Mutex};
 use std::thread;
 use std::time;
 
+use anyhow::{anyhow, Result};
 use ignore::overrides::OverrideBuilder;
 use ignore::{self, WalkBuilder};
 use regex::bytes::Regex;
 
+use crate::error::print_error;
+use crate::exec;
+use crate::exit_codes::{merge_exitcodes, ExitCode};
+use crate::filesystem;
+use crate::options::Options;
+use crate::output;
+
 /// The receiver thread can either be buffering results or directly streaming to the console.
 enum ReceiverMode {
     /// Receiver is still buffering in order to sort the results, if the search finishes fast
@@ -44,12 +38,15 @@ pub enum WorkerResult {
     Error(ignore::Error),
 }
 
+/// Maximum size of the output buffer before flushing results to the console
+pub const MAX_BUFFER_LENGTH: usize = 1000;
+
 /// Recursively scan the given search path for files / pathnames matching the pattern.
 ///
 /// If the `--exec` argument was supplied, this will create a thread pool for executing
 /// jobs in parallel from a given command line and the discovered paths. Otherwise, each
 /// path will simply be written to standard output.
-pub fn scan(path_vec: &[PathBuf], pattern: Arc<Regex>, config: Arc<FdOptions>) -> ExitCode {
+pub fn scan(path_vec: &[PathBuf], pattern: Arc<Regex>, config: Arc<Options>) -> Result<ExitCode> {
     let mut path_iter = path_vec.iter();
     let first_path_buf = path_iter
         .next()
@@ -59,14 +56,13 @@ pub fn scan(path_vec: &[PathBuf], pattern: Arc<Regex>, config: Arc<FdOptions>) -
     let mut override_builder = OverrideBuilder::new(first_path_buf.as_path());
 
     for pattern in &config.exclude_patterns {
-        let res = override_builder.add(pattern);
-        if res.is_err() {
-            print_error_and_exit!("Malformed exclude pattern '{}'", pattern);
-        }
+        override_builder
+            .add(pattern)
+            .map_err(|e| anyhow!("Malformed exclude pattern: {}", e))?;
     }
-    let overrides = override_builder.build().unwrap_or_else(|_| {
-        print_error_and_exit!("Mismatch in exclude patterns");
-    });
+    let overrides = override_builder
+        .build()
+        .map_err(|_| anyhow!("Mismatch in exclude patterns"))?;
 
     let mut walker = WalkBuilder::new(first_path_buf.as_path());
     walker
@@ -78,25 +74,51 @@ pub fn scan(path_vec: &[PathBuf], pattern: Arc<Regex>, config: Arc<FdOptions>) -
         .git_exclude(config.read_vcsignore)
         .overrides(overrides)
         .follow_links(config.follow_links)
+        // No need to check for supported platforms, option is unavailable on unsupported ones
+        .same_file_system(config.one_file_system)
         .max_depth(config.max_depth);
 
     if config.read_fdignore {
         walker.add_custom_ignore_filename(".fdignore");
     }
 
+    if config.read_global_ignore {
+        #[cfg(target_os = "macos")]
+        let config_dir_op = std::env::var_os("XDG_CONFIG_HOME")
+            .map(PathBuf::from)
+            .filter(|p| p.is_absolute())
+            .or_else(|| dirs::home_dir().map(|d| d.join(".config")));
+
+        #[cfg(not(target_os = "macos"))]
+        let config_dir_op = dirs::config_dir();
+
+        if let Some(global_ignore_file) = config_dir_op
+            .map(|p| p.join("fd").join("ignore"))
+            .filter(|p| p.is_file())
+        {
+            let result = walker.add_ignore(global_ignore_file);
+            match result {
+                Some(ignore::Error::Partial(_)) => (),
+                Some(err) => {
+                    print_error(format!(
+                        "Malformed pattern in global ignore file. {}.",
+                        err.to_string()
+                    ));
+                }
+                None => (),
+            }
+        }
+    }
+
     for ignore_file in &config.ignore_files {
         let result = walker.add_ignore(ignore_file);
         match result {
             Some(ignore::Error::Partial(_)) => (),
             Some(err) => {
-                print_error!(
-                    "{}",
-                    format!(
-                        "Malformed pattern in custom ignore file '{}': {}.",
-                        ignore_file.to_string_lossy(),
-                        err.description()
-                    )
-                );
+                print_error(format!(
+                    "Malformed pattern in custom ignore file. {}.",
+                    err.to_string()
+                ));
             }
             None => (),
         }
@@ -132,14 +154,14 @@ pub fn scan(path_vec: &[PathBuf], pattern: Arc<Regex>, config: Arc<FdOptions>) -
     let exit_code = receiver_thread.join().unwrap();
 
     if wants_to_quit.load(Ordering::Relaxed) {
-        process::exit(ExitCode::KilledBySigint.into());
+        Ok(ExitCode::KilledBySigint)
+    } else {
+        Ok(exit_code)
     }
-
-    exit_code
 }
 
 fn spawn_receiver(
-    config: &Arc<FdOptions>,
+    config: &Arc<Options>,
     wants_to_quit: &Arc<AtomicBool>,
     rx: Receiver<WorkerResult>,
 ) -> thread::JoinHandle<ExitCode> {
@@ -175,11 +197,12 @@ fn spawn_receiver(
                 }
 
                 // Wait for all threads to exit before exiting the program.
+                let mut results: Vec<ExitCode> = Vec::new();
                 for h in handles {
-                    h.join().unwrap();
+                    results.push(h.join().unwrap());
                 }
 
-                ExitCode::Success
+                merge_exitcodes(&results)
             }
         } else {
             let start = time::Instant::now();
@@ -197,6 +220,8 @@ fn spawn_receiver(
             let stdout = io::stdout();
             let mut stdout = stdout.lock();
 
+            let mut num_results = 0;
+
             for worker_result in rx {
                 match worker_result {
                     WorkerResult::Entry(value) => {
@@ -227,13 +252,21 @@ fn spawn_receiver(
                                 output::print_entry(&mut stdout, &value, &config, &wants_to_quit);
                             }
                         }
+
+                        num_results += 1;
                     }
                     WorkerResult::Error(err) => {
                         if show_filesystem_errors {
-                            print_error!("{}", err);
+                            print_error(err.to_string());
                         }
                     }
                 }
+
+                if let Some(max_results) = config.max_results {
+                    if num_results >= max_results {
+                        break;
+                    }
+                }
             }
 
             // If we have finished fast enough (faster than max_buffer_time), we haven't streamed
@@ -250,8 +283,45 @@ fn spawn_receiver(
     })
 }
 
+pub enum DirEntry {
+    Normal(ignore::DirEntry),
+    BrokenSymlink(PathBuf),
+}
+
+impl DirEntry {
+    pub fn path(&self) -> &Path {
+        match self {
+            DirEntry::Normal(e) => e.path(),
+            DirEntry::BrokenSymlink(pathbuf) => pathbuf.as_path(),
+        }
+    }
+
+    pub fn file_type(&self) -> Option<FileType> {
+        match self {
+            DirEntry::Normal(e) => e.file_type(),
+            DirEntry::BrokenSymlink(pathbuf) => {
+                pathbuf.symlink_metadata().map(|m| m.file_type()).ok()
+            }
+        }
+    }
+
+    pub fn metadata(&self) -> Option<Metadata> {
+        match self {
+            DirEntry::Normal(e) => e.metadata().ok(),
+            DirEntry::BrokenSymlink(_) => None,
+        }
+    }
+
+    pub fn depth(&self) -> Option<usize> {
+        match self {
+            DirEntry::Normal(e) => Some(e.depth()),
+            DirEntry::BrokenSymlink(_) => None,
+        }
+    }
+}
+
 fn spawn_senders(
-    config: &Arc<FdOptions>,
+    config: &Arc<Options>,
     wants_to_quit: &Arc<AtomicBool>,
     pattern: Arc<Regex>,
     parallel_walker: ignore::WalkParallel,
@@ -269,27 +339,53 @@ fn spawn_senders(
             }
 
             let entry = match entry_o {
-                Ok(e) => e,
+                Ok(ref e) if e.depth() == 0 => {
+                    // Skip the root directory entry.
+                    return ignore::WalkState::Continue;
+                }
+                Ok(e) => DirEntry::Normal(e),
+                Err(ignore::Error::WithPath {
+                    path,
+                    err: inner_err,
+                }) => match inner_err.as_ref() {
+                    ignore::Error::Io(io_error)
+                        if io_error.kind() == io::ErrorKind::NotFound
+                            && path
+                                .symlink_metadata()
+                                .ok()
+                                .map_or(false, |m| m.file_type().is_symlink()) =>
+                    {
+                        DirEntry::BrokenSymlink(path)
+                    }
+                    _ => {
+                        tx_thread
+                            .send(WorkerResult::Error(ignore::Error::WithPath {
+                                path,
+                                err: inner_err,
+                            }))
+                            .unwrap();
+                        return ignore::WalkState::Continue;
+                    }
+                },
                 Err(err) => {
                     tx_thread.send(WorkerResult::Error(err)).unwrap();
                     return ignore::WalkState::Continue;
                 }
             };
 
-            if entry.depth() == 0 {
-                return ignore::WalkState::Continue;
+            if let Some(min_depth) = config.min_depth {
+                if entry.depth().map_or(true, |d| d < min_depth) {
+                    return ignore::WalkState::Continue;
+                }
             }
 
             // Check the name first, since it doesn't require metadata
             let entry_path = entry.path();
 
             let search_str: Cow<OsStr> = if config.search_full_path {
-                match fshelper::path_absolute_form(entry_path) {
-                    Ok(path_abs_buf) => Cow::Owned(path_abs_buf.as_os_str().to_os_string()),
-                    Err(_) => {
-                        print_error_and_exit!("Unable to retrieve absolute path.");
-                    }
-                }
+                let path_abs_buf = filesystem::path_absolute_form(entry_path)
+                    .expect("Retrieving absolute path succeeds");
+                Cow::Owned(path_abs_buf.as_os_str().to_os_string())
             } else {
                 match entry_path.file_name() {
                     Some(filename) => Cow::Borrowed(filename),
@@ -301,14 +397,14 @@ fn spawn_senders(
                 }
             };
 
-            if !pattern.is_match(&osstr_to_bytes(search_str.as_ref())) {
+            if !pattern.is_match(&filesystem::osstr_to_bytes(search_str.as_ref())) {
                 return ignore::WalkState::Continue;
             }
 
             // Filter out unwanted extensions.
             if let Some(ref exts_regex) = config.extensions {
                 if let Some(path_str) = entry_path.file_name() {
-                    if !exts_regex.is_match(&osstr_to_bytes(path_str)) {
+                    if !exts_regex.is_match(&filesystem::osstr_to_bytes(path_str)) {
                         return ignore::WalkState::Continue;
                     }
                 } else {
@@ -322,26 +418,40 @@ fn spawn_senders(
                     if (!file_types.files && entry_type.is_file())
                         || (!file_types.directories && entry_type.is_dir())
                         || (!file_types.symlinks && entry_type.is_symlink())
+                        || (!file_types.sockets && filesystem::is_socket(entry_type))
+                        || (!file_types.pipes && filesystem::is_pipe(entry_type))
                         || (file_types.executables_only
                             && !entry
                                 .metadata()
-                                .map(|m| fshelper::is_executable(&m))
+                                .map(|m| filesystem::is_executable(&m))
                                 .unwrap_or(false))
-                        || (file_types.empty_only && !fshelper::is_empty(&entry))
+                        || (file_types.empty_only && !filesystem::is_empty(&entry))
+                        || !(entry_type.is_file()
+                            || entry_type.is_dir()
+                            || entry_type.is_symlink()
+                            || filesystem::is_socket(entry_type)
+                            || filesystem::is_pipe(entry_type))
                     {
                         return ignore::WalkState::Continue;
-                    } else if !(entry_type.is_file()
-                        || entry_type.is_dir()
-                        || entry_type.is_symlink())
-                    {
-                        // This is probably a block device, char device, fifo or socket. Skip it.
-                        return ignore::WalkState::Continue;
                     }
                 } else {
                     return ignore::WalkState::Continue;
                 }
             }
 
+            #[cfg(unix)]
+            {
+                if let Some(ref owner_constraint) = config.owner_constraint {
+                    if let Ok(ref metadata) = entry_path.metadata() {
+                        if !owner_constraint.matches(&metadata) {
+                            return ignore::WalkState::Continue;
+                        }
+                    } else {
+                        return ignore::WalkState::Continue;
+                    }
+                }
+            }
+
             // Filter out unwanted sizes if it is a file and we have been given size constraints.
             if !config.size_constraints.is_empty() {
                 if entry_path.is_file() {
@@ -378,10 +488,11 @@ fn spawn_senders(
                 }
             }
 
-            // TODO: take care of the unwrap call
-            tx_thread
-                .send(WorkerResult::Entry(entry_path.to_owned()))
-                .unwrap();
+            let send_result = tx_thread.send(WorkerResult::Entry(entry_path.to_owned()));
+
+            if send_result.is_err() {
+                return ignore::WalkState::Quit;
+            }
 
             ignore::WalkState::Continue
         })
